package main

import (
	"dwelling-upload/internal/configuration"
	"dwelling-upload/internal/http"
	"dwelling-upload/pkg/logging"
	"dwelling-upload/pkg/utils"
	"dwelling-upload/pkg/watcher"
	"flag"
	"fmt"
	"log"
	"os"
	"os/signal"
	"syscall"
)

var configPath *string = flag.String("conf", "config.yaml", "path to configuration file")
var showVersion *bool = flag.Bool("v", false, "show version")

var version string

func main() {
	flag.Parse()

	if *showVersion {
		fmt.Println("dwelling-upload Ver. ", version, "\nCopyright (c) 2022,2023 Alexander \"Arav\" Andreev <me@arav.su>")
		return
	}

	config, err := configuration.LoadConfiguration(*configPath)
	if err != nil {
		log.Fatalln(err)
	}

	defer func() {
		if typ, addr := config.SplitNetworkAddress(); typ == "unix" {
			os.Remove(addr)
		}
	}()

	logErr, err := logging.New(config.Log.Error)
	if err != nil {
		log.Fatalln("error logger:", err)
	}
	defer logErr.Close()

	logFile, err := logging.New(config.Log.File)
	if err != nil {
		log.Fatalln("file logger:", err)
	}
	defer logFile.Close()

	watcha, err := watcher.NewInotifyWatcher()
	if err != nil {
		logErr.Fatalln(err)
	}
	defer watcha.Close()

	if err := watcha.AddWatch(config.Uploads.Directory, watcher.CrDelMask); err != nil {
		logErr.Fatalln(err)
	}

	uploadDirNotify := make(chan uint32)
	uploadDirSize, err := utils.DirectorySize(config.Uploads.Directory)
	if err != nil {
		logErr.Fatalf("failed to get initial size of %s: %s", config.Uploads.Directory, err)
	}

	watcha.WatchForMask(uploadDirNotify, watcher.CrDelMask)

	hand := http.NewUploadHandlers(config, logErr, logFile, &uploadDirSize)
	srv := http.NewHttpServer()

	srv.SetNotFoundHandler(http.NotFound)

	srv.ServeStatic("/assets/*filepath", hand.AssetsFS())
	srv.GET("/robots.txt", http.RobotsTxt)
	srv.GET("/favicon.svg", http.Favicon)

	srv.GET("/", hand.Index)
	srv.POST("/", hand.Upload)
	srv.POST("/delete", hand.Delete)
	srv.GET("/f/:hash/:name", hand.Download)
	srv.DELETE("/:hash", hand.Delete)

	if err := srv.Start(config.SplitNetworkAddress()); err != nil {
		logErr.Fatalln("failed to start a server:", err)
	}

	doneSignal := make(chan os.Signal, 1)
	signal.Notify(doneSignal, os.Interrupt, syscall.SIGINT, syscall.SIGTERM)
	logReopenSignal := make(chan os.Signal, 1)
	signal.Notify(logReopenSignal, syscall.SIGHUP)

	go func() {
		for {
			select {
			case <-logReopenSignal:
				logErr.Reopen(config.Log.Error)
				logFile.Reopen(config.Log.File)
			case <-uploadDirNotify:
				uploadDirSize, err = utils.DirectorySize(config.Uploads.Directory)
				if err != nil {
					logErr.Println("failed to get uploads directory size:", err)
				}
			}
		}
	}()

	<-doneSignal

	if err := srv.Stop(); err != nil {
		logErr.Fatalln("failed to properly shutdown a server:", err)
	}
}
